﻿using MyFirstApp.Data;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Runtime.Serialization;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Xml;
using System.Xml.Linq;
using Windows.ApplicationModel.DataTransfer;
using Windows.Data.Xml.Dom;
using Windows.Foundation;
using Windows.Networking.Connectivity;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.UI.Notifications;
using Windows.UI.Popups;
using Windows.UI.StartScreen;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Media;
using Windows.Web.Syndication;

namespace MyFirstApp
{
	public static class Utilities
	{
		public static string ExtractText(this string text)
		{
			return !string.IsNullOrEmpty(text) ? WebUtility.HtmlDecode(text) : string.Empty;
		}

		private static Regex _cleanerRegex = new Regex(@"</?\w+((\s+\w+(\s*=\s*(?:"".*?""|'.*?'|[^'"">\s]+))?)+\s*|\s*)/?>");
		/// <summary>
		/// ENG: removes HTML and get a plain text version
		/// ITA: Rimuove l'HTML ed ottengo una versione plain text
		/// </summary>
		/// <param name="html"></param>
		/// <returns></returns>
		public static string ToPlainText(this string html)
		{
			return _cleanerRegex.Replace(html, String.Empty).Trim();
		}
	}

	public class ServiceManager
	{
		internal static async Task<SampleDataGroup> LoadGroupAsync(Uri url, string title)
		{
			string cacheKey = string.Join("", url.ToString().ToCharArray().Where(x => Char.IsDigit(x)).ToArray());

			return await ServiceManager.ExecuteCachedCallAsync(async () =>
			{
				var group = new SampleDataGroup(url.ToString(), title, null, null, title);

				try
				{
					var client = new HttpClient();
					client.DefaultRequestHeaders.Add("user-agent", "msie10");
					var xml = await client.GetStringAsync(url);
					//var doc = XDocument.Load(url.ToString());
					var doc = XDocument.Parse(xml);
					var items = doc.Descendants("item").Select(x =>
						new SampleDataItem(x.Element("guid").Value, x.Element("title").Value, null, GetImage(x.Element("description").Value), x.Element("description").Value, x.Element("description").Value, group));

					foreach (var item in items)
					{
						group.Items.Add(item);
					}
				}
				catch (Exception e)
				{
					Debug.WriteLine(e.Message);
				}
				return group;
			}, cacheKey, false);
		}

		private static string GetImage(string text)
		{
			if (string.IsNullOrEmpty(text))
				return null;

			// ENG: let's try with links pointing to images
			// then we'll try with images directly
			// ITA: Cerco prima i link che puntano ad immagini
			// Poi tento con le immagini direttamente
			string[] patterns = new[] {
                "href\\s*=\\s*(?:\"(?<1>[^\"]*)\"|(?<1>\\S+))",
                "src\\s*=\\s*(?:\"(?<1>[^\"]*)\"|(?<1>\\S+))"
            };

			foreach (string p in patterns)
			{
				var s = Regex.Matches(text,
						p,
						RegexOptions.None)
					.Cast<Match>()
					.Where(m =>
					{
						Uri url;
						if (Uri.TryCreate(m.Groups[1].Value, UriKind.Absolute, out url))
						{
							string ext = Path.GetExtension(url.AbsolutePath).ToLower();
							if (ext == ".png" || ext == ".jpg") return true;
						}
						return false;
					})
					.Select(m => m.Groups[1].Value)
					.FirstOrDefault();
				if (s != null)
					return s;
			}

			return null;
		}

		public static bool CheckIsNetworkAvailable()
		{
			var connectionProfile = NetworkInformation.GetInternetConnectionProfile();

			return connectionProfile != null &&
				connectionProfile.GetNetworkConnectivityLevel() == NetworkConnectivityLevel.InternetAccess;
		}

		public static async Task ClearCache(string cacheKey)
		{
			string fileName = cacheKey + ".cache";


			IStorageFile cacheFile = await Persister.GetFile(fileName, false);
			if (cacheFile != null)
			{
				// il file è troppo vecchio, lo eliminiamo
#pragma warning disable 4014
				cacheFile.DeleteAsync();
#pragma warning restore 4014
			}
		}

		public static async Task<T> ExecuteCachedCallAsync<T>(Func<Task<T>> loadTask, string cacheKey, bool clearCache)
		{
			string fileName = cacheKey + ".cache";

			IStorageFile cacheFile = await Persister.GetFile(fileName, createIfNotFound: false);

			// no cache
			if (clearCache)
				cacheKey = null;

			if (!CheckIsNetworkAvailable())
			{
				// rete non disponibile, c'è un file recente?
				if (cacheFile == null)
					throw new WebException();

				var properties = (cacheFile != null) ? await cacheFile.GetBasicPropertiesAsync() : null;
				if ((DateTime.Now - properties.DateModified) > TimeSpan.FromDays(3))
				{
					// il file è troppo vecchio, lo eliminiamo
#pragma warning disable 4014
					cacheFile.DeleteAsync();
#pragma warning restore 4014
					throw new WebException();
				}

				// il file è abbastanza recente, lo restituiamo
				return await Persister.ReadFile<T>(cacheFile);
			}
			else
			{
				// Recupero le informazioni sul file
				var properties = (cacheFile != null) ? await cacheFile.GetBasicPropertiesAsync() : null;

				// rete disponibile, ma se il file è molto recente lo riutilizzo
				if (cacheFile != null && (DateTime.Now - properties.DateModified) < TimeSpan.FromMinutes(5))
				{
					var item2 = await Persister.ReadFile<T>(cacheFile);
					if (item2 != null)
						return item2;
				}

				// se arriviamo qui o non c'è in cache, oppure non è stato letto correttamente
				if (cacheFile != null)
				{
					await cacheFile.DeleteAsync();
				}

				// recupero dalla rete e salvo il contenuto sul file per il prossimo caricamento
				var item = await loadTask();

				cacheFile = await Persister.GetFile(fileName, createIfNotFound: true);
#pragma warning disable 4014
				await Persister.WriteFile<T>(item, cacheFile);
#pragma warning restore 4014

				return item;

			}
		}
	}

	public static class PinHelper
	{
		public static async Task<bool> IsPinned(SampleDataItem content)
		{
			var results = await SecondaryTile.FindAllAsync();
			return results.Count(x => x.TileId == GetPinId(content)) > 0;
		}

		internal static string GetPinId(SampleDataItem content)
		{
			return content.UniqueId.GetHashCode().ToString();
		}

		public async static Task RemovePin(SampleDataItem content)
		{
			var results = await SecondaryTile.FindAllAsync();
			var tile = results.FirstOrDefault(x => x.TileId == GetPinId(content));
			if (tile != null)
			{
				await tile.RequestDeleteAsync();
			}
		}

		public async static Task<bool> Pin(SampleDataItem content, FrameworkElement anchorElement)
		{
			SecondaryTile tile = new SecondaryTile(GetPinId(content), content.Title, content.Title, content.UniqueId, TileOptions.ShowNameOnLogo | TileOptions.ShowNameOnWideLogo | TileOptions.CopyOnDeployment, new System.Uri("ms-appx:///Assets/Logo.png"), new System.Uri("ms-appx:///Assets/WideLogo.png"));
			tile.ForegroundText = ForegroundText.Dark;

			// ITA: creazione tile
			bool success = await tile.RequestCreateForSelectionAsync(GetElementRect(anchorElement), Placement.Left);
			if (success)
			{
				// ITA: aggiorno la tile appena creata
				TileHelper.UpdateSecondaryTile(content, tile.WideLogo, tile.Logo);

				await new MessageDialog("Pin OK!").ShowAsync();
				return true;
			}
			else
			{
				// error
				await new MessageDialog("Pin annullato").ShowAsync();
				return false;
			}
		}


		private static Rect GetElementRect(FrameworkElement element)
		{
			GeneralTransform buttonTransform = element.TransformToVisual(null);
			Point point = buttonTransform.TransformPoint(new Point());
			return new Rect(point, new Size(0, element.ActualHeight));
		}
	}


	public static class TileHelper
	{

		public static void UpdatePrimaryTile(SampleDataItem content, Uri wideLogo = null, Uri logo = null)
		{
			TileUpdateManager.CreateTileUpdaterForApplication().EnableNotificationQueue(true);

			if (logo == null)
				logo = new System.Uri("ms-appx:///Assets/Logo.png");

			if (wideLogo == null)
				wideLogo = new System.Uri("ms-appx:///Assets/WideLogo.png");

			// ITA: recuperiamo l'XML
			XmlDocument xmlDoc = GetTileXml(content, wideLogo, logo);

			try
			{
				var tile = new TileNotification(xmlDoc);

				// ITA: impostiamo la proprietà Tag, così la tile sarà identificata univocamente sarà possibile ciclare tra le ultime n tile
				tile.Tag = content.UniqueId.GetHashCode().ToString();

				// ITA: aggiorniamo la tile
				TileUpdateManager.CreateTileUpdaterForApplication().Update(tile);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("Tile: " + ex.Message);
			}
		}

		public static void UpdateSecondaryTile(SampleDataItem content, Uri wideLogo, Uri logo)
		{
			string pinId = PinHelper.GetPinId(content);

			// ENG: let's check if the tile exists
			// ITA: secondary tile, controlliamo prima che esista
			if (SecondaryTile.Exists(pinId))
			{
				// ENG: get the XML
				// ITA: recuperiamo l'XML
				XmlDocument xmlDoc = GetTileXml(content, wideLogo, logo);

				// ENG: let's update the tile
				// ITA: aggiorniamo la tile
				var secTile = TileUpdateManager.CreateTileUpdaterForSecondaryTile(pinId);
				secTile.Update(new TileNotification(xmlDoc));
			}
		}

		private static XmlDocument GetTileXml(SampleDataItem content, Uri wideLogo, Uri logo)
		{
			// ENG: the tile is created using LINQ TO XML
			// ITA: componiamo la tile con LINQ to XML
			var tileXml = new XDocument(new XElement("tile",
							new XElement("visual",
								new XElement("binding",
									new XAttribute("template", "TileWideImageAndText01"),
										new XElement("image", new XAttribute("id", "1"), new XAttribute("src", content.ImageUrl ?? wideLogo.ToString())),
										new XElement("text", new XAttribute("id", "1"), content.Title),
										new XElement("text", new XAttribute("id", "2"), content.Subtitle)
								),
								new XElement("binding",
									new XAttribute("template", "TileSquarePeekImageAndText04"),
										new XElement("image", new XAttribute("id", "1"), new XAttribute("src", content.ImageUrl ?? logo.ToString())),
										new XElement("text", new XAttribute("id", "1"), content.Title),
										new XElement("text", new XAttribute("id", "2"), content.Subtitle)
										)
								)
							));

			var xmlDoc = new XmlDocument();
			xmlDoc.LoadXml("<?xml version=\"1.0\" encoding=\"utf-8\" ?>" + tileXml.ToString());
			return xmlDoc;
		}
	}

	public class Persister
	{
		private static List<Type> knownTypes = new List<Type>();

		static Persister()
		{
			knownTypes.Add(typeof(String[]));
			knownTypes.Add(typeof(Uri[]));
			knownTypes.Add(typeof(String[][]));
			knownTypes.Add(typeof(Int32[]));
			knownTypes.Add(typeof(Int16[]));
			knownTypes.Add(typeof(Byte[]));
			knownTypes.Add(typeof(Single[]));
			knownTypes.Add(typeof(Double[]));
			knownTypes.Add(typeof(DateTime));
		}

		public static async Task<IStorageFile> GetFile(string fileName, bool createIfNotFound)
		{
			StorageFolder folder = ApplicationData.Current.LocalFolder;

			try
			{
				return await folder.GetFileAsync(fileName);
			}
			catch (FileNotFoundException)
			{ }

			// se siamo arrivati fin qui, il file non esiste
			if (createIfNotFound)
				return await folder.CreateFileAsync(fileName);

			return null;
		}

		public static void WriteObject<T>(Stream stream, T obj)
		{
			DataContractSerializer serializer = new DataContractSerializer(typeof(T), new DataContractSerializerSettings()
			{
				KnownTypes = knownTypes,
				DataContractResolver = new GenericTypeResolver()
			});

			serializer.WriteObject(stream, obj);
		}

		public static T ReadObject<T>(Stream stream)
		{
			try
			{
				DataContractSerializer serializer = new DataContractSerializer(typeof(T), new DataContractSerializerSettings()
				{
					KnownTypes = knownTypes,
					DataContractResolver = new GenericTypeResolver()
				});

				return (T)serializer.ReadObject(stream);
			}
			catch
			{

				return default(T);
			}
		}

		internal async static Task WriteFile<T>(T item, IStorageFile cacheFile)
		{
			try
			{
				if (cacheFile == null) return;

				var ouStream = await cacheFile.OpenStreamForWriteAsync();
				// Deserializzo dall'XML
				using (Stream stream = ouStream.AsOutputStream().AsStreamForWrite())
				{
					WriteObject<T>(stream, item);
				}
			}
			catch (Exception e)
			{
				// In caso di errore, ignoro
				Debug.WriteLine(e);
			}

			return;
		}

		internal async static Task<T> ReadFile<T>(IStorageFile cacheFile)
		{
			try
			{
				if (cacheFile == null) return default(T);
				IInputStream inStream = await cacheFile.OpenSequentialReadAsync();

				// Deserializzo dall'XML
				using (Stream stream = inStream.AsStreamForRead())
				{
					return ReadObject<T>(stream);
				}
			}
			catch (Exception e)
			{
				// In caso di errore, ignoro
				Debug.WriteLine(e);
			}

			return default(T);
		}
	}

	public class GenericTypeResolver : DataContractResolver
	{
		private const string RESOLVER_NAMESPACE = "http://www.ASPItalia.com/WinRT/genericTypeResolver";
		public override Type ResolveName(string typeName, string typeNamespace, Type declaredType, DataContractResolver knownTypeResolver)
		{
			if (typeNamespace == RESOLVER_NAMESPACE)
			{
				return Type.GetType(typeName);
			}
			else
			{
				return knownTypeResolver.ResolveName(typeName, typeNamespace, declaredType, null);
			}
		}

		public override bool TryResolveType(Type type, Type declaredType, DataContractResolver knownTypeResolver, out XmlDictionaryString typeName, out XmlDictionaryString typeNamespace)
		{
			// try resolving with knowtypes
			if (!knownTypeResolver.TryResolveType(type, declaredType, null, out typeName, out typeNamespace))
			{
				XmlDictionary dictionary = new XmlDictionary();
				typeName = dictionary.Add(type.AssemblyQualifiedName);
				typeNamespace = dictionary.Add(RESOLVER_NAMESPACE);
			}

			return true;
		}
	}
}